CREATE OR REPLACE  PROCEDURE "AUTH"."AUTH_CONNECTION"  (p_return 
    OUT varchar, p_validate in varchar default 'FALSE' )
AS
 /*
  AUTH_CONNECTION - This procedure controls database authentication. It is called from the
  SYS.AUTH_CONNECTION_TRIG trigger which fires after logon to the database  Access is configured
  in the auth.auth_control table and is able to test the user's name, roles, machine name
  executing program name, excuting module name (dependeding on the client type and version),
  and test if the connection is from a database link. The value test is not case sensitive.
  To gain access a user must meet an AUTORIZED condition with a permission_level greater than
  the highest permission_level for any UNAUTORIZED condition that user's connection meets.
  The action taken by the procedure is also configured in the auth_control.  See defination
  below outlines the options:

  auth.auth_control Table Defination
    enabled          -- 'Y' Enables (default) and 'N' disables the condition.
    condition        -- ROLE,MACHINE,PROGRAM,MODULE,LINK,USERNAME, or OSUSER
                        and the compaired value. The % wildcard is allowed for the values
    permission_level -- Higher Level = Hight Importance, (default is 1).
                        Access is granted if the greatest UNAUTHORIZED level is
                        less then the greatest AUTORIZED level.
    permission       -- AUTORIZED or UNAUTHORIZED (default) condition
    action           -- NONE = No Action except when unless email_notice = Y
                        LOG = Log access to auth_log and send email if email_notice = Y, Does not Block session.
                        TRACE  = Start a SQL Trace and send email if email_notice = Y
                        RECORD = Start Trace, convert trace file to SQL RECORDING and send email if email_notice = Y
                        BLOCK = Block connection and send email if email_notice = Y
    email_notice     -- 'Y' Enables and 'N' disables (default) email notification when the condition evaluates ture
                        and has the highest permission_level.  'S' or "Sticky" enables email notification when the
                        condition evaluates ture even if the condition does not have the highest permission_level.

  Email notices are sent imediately if a condition is met where an email notice is requested with
  email_notice = Y. The email will be sent to each email_address in the auth.auth_email table where
  the enabled column = Y.

  When p_validate = 'FALSE' The call is from the AUTH_CONNECTION_TRIG and the apporpriate action will be taken
  but when p_validate = 'TRUE' the call is from a tabel trigger and will only validate the condition.  The
  reformated condition and any error message will be returned in p_return.

  -- Programmer: Doug Jackson
  -- October 8, 2004
  -- Oct 25, 2004 -- Added ability to use NOT like.
  -- Nov 1,  2004 -- Fixed ROLE check from returning multiple roles
  -- Nov 2,  2004 -- Modified to handel longer strings
  -- Nov 4,  2004 -- Modified Program to report 'Unknown' when null
  -- Nov 10, 2004
     Modified email notice to 'Y','N', or 'S' for "sticky" Email will only be sent if 'Y' for the
     condition with the highest permission_level unless 'S' as set for another valid condition.
     -- Structure Changes:
     alter table auth.auth_control drop constraint auth_control_email_notice_ck;
     alter table auth.auth_control add constraint auth_control_email_notice_ck check (email_notice in ('Y','N','S'));
     alter table auth.auth_control modify (email_notice char(1)  default 'N');
     -- As SYSDBA
     grant execute on sys.dbms_system to auth;

  Sample auth_control records:
  -- Authorize Database Links
  insert into auth.auth_control (condition,permission,permission_level,action,email_notice)
    values ('LINK like oracle@rs%','AUTHORIZED',2,'NONE','N');
  -- Logging
  insert into auth.auth_control (condition,permission,permission_level,action,email_notice)
    values ('PROGRAM like % and (ROLE NOT LIKE DBA and MACHINE NOT LIKE rsprod%)','AUTHORIZED',1,'LOG','N');
  -- Blocking
  insert into auth.auth_control (condition,permission,permission_level,action,email_notice)
    values ('PROGRAM LIKE %TOAD% and (ROLE NOT LIKE DBA and MACHINE NOT LIKE rsprod%)','UNAUTHORIZED',1,'BLOCK','Y');
  insert into auth.auth_control (condition,permission,permission_level,action,email_notice)
    values ('PROGRAM LIKE %SQL% and (ROLE NOT LIKE DBA and MACHINE NOT LIKE rsprod%)','UNAUTHORIZED',1,'BLOCK','Y');
  commit;
  -- Authorize SU for DBA
  insert into auth.auth_control (condition,permission,permission_level,action,description)
    values ('OSUSER like OKE03910','AUTHORIZED',9,'NONE','Doug Jackson');
  insert into auth.auth_control (condition,permission,permission_level,action,description)
    values ('OSUSER like OKE03789','AUTHORIZED',9,'NONE','Brian Cleavelin');

 */

  pragma autonomous_transaction;
  v_sid                         number;
  v_serial                      number;
  v_spid                        number;
  v_logon_time                  date;
  v_username                    varchar2(30);
  v_dba                         varchar2(3);
  v_link                        varchar2(3);
  v_osuser                      varchar2(30);
  v_program                     varchar2(48);
  v_module                      varchar2(48);
  v_machine                     varchar2(64);
  v_check                       varchar2(5);
  v_unauthorized_condition      auth.auth_control.condition%TYPE;
  v_unauthorized_condition_eval auth.auth_log.condition_eval%TYPE;
  v_unauthorized_level          auth.auth_control.permission_level%TYPE;
  v_unauthorized_action         auth.auth_control.action%TYPE;
  v_unauthorized_email_notice   auth.auth_control.email_notice%TYPE;
  v_block_msg                   auth.auth_control.block_msg%TYPE;
  v_authorized_condition        auth.auth_control.condition%TYPE;
  v_authorized_condition_eval   auth.auth_log.condition_eval%TYPE;
  v_authorized_level            auth.auth_control.permission_level%TYPE;
  v_authorized_action           auth.auth_control.action%TYPE;
  v_authorized_email_notice     auth.auth_control.email_notice%TYPE;
  v_action                      auth.auth_control.action%TYPE;
  v_method                      varchar2(30);
  v_value                       varchar2(30);
  v_log                         varchar2(3);
  v_log_id                      number;
  v_sticky_email                varchar2(3);
  v_email                       auth.auth_control.email_notice%TYPE;
  v_trace                       varchar2(10);
  v_trace_path_file             varchar2(200);
  v_instance_name               varchar2(16);
  v_host_name                   varchar2(64);
  v_body                        varchar2(1000);
  v_count                       number;
  v_loc                         number;
  v_beg                         number;
  v_length                      number;
  v_condition                   auth.auth_control.condition%TYPE;
  v_condition_eval              auth.auth_log.condition_eval%TYPE;
  v_single_condition            varchar2(1000);
  v_operand                     varchar2(10);
  v_error_condition             varchar2(1000);
  v_sql                         varchar2(1000);
begin
  -- Get Instance and Host names
  select instance_name, host_name into v_instance_name, v_host_name from v$instance;
  -- Get the Session ID
  select distinct sid  into v_sid from sys.v_$mystat;
  -- Get the contents of v$session. Note: "module" does not get populated on older client versions.
  select replace(machine,chr(0),''), username, osuser, nvl(program,'Unknown'), module into v_machine, v_username, v_osuser, v_program, v_module
    from sys.v_$session where sid = v_sid;
  -- Get DBO info
  select max(ans) into v_dba from
    (select 'YES' as ans from v$pwfile_users where username = v_username
      union
     select 'YES' as ans from dba_role_privs where granted_role = 'DBA' and grantee = v_username
      union
     select 'NO' as ans from dual);

  -- Step throught each enabled condition ordered by permission_level asc and test each condition
  -- remembering the highest level qualifing UNAUTHORIZED and AUTHORIZED condition and any qualifying
  -- condition that requires loging or email notice.

  -- Clear Flags for loging and sending email notice
  v_log   := 'NO';
  v_sticky_email := 'NO';
  v_trace := 'NO';
  v_authorized_level := 0;
  v_unauthorized_level := 0;
  v_block_msg := 'Access Blocked';

  -- Delete any expired auth_control records
  begin
    delete auth.auth_control where expires < sysdate;
    commit;
  exception when others then
    null;
  end;

  for auth_control in (select permission_level, permission, condition, action, email_notice, block_msg
                       from auth.auth_control where enabled='Y'
                       order by permission_level asc  ) loop

    -- Sample condition: 'PROGRAM like %TOAD% and (MACHINE like %78WXVGW-OKE or USERNAME like TEST)'
    -- Identify each pair, ie PROGRAM like %TOAD% and reduce to TRUE, FALSE, or ERROR
    v_condition := ' '|| replace( replace( upper(auth_control.condition),'(',' ( ') ,')',' ) ') || ' ';
    while length(v_condition) > length(replace(v_condition,'  ',' ')) loop
      v_condition := replace(v_condition,'  ',' ');
    end loop;
    v_error_condition := lower(v_condition);

    -- Step through each clause in the condition
    v_count := 1;
    loop
      v_loc := instr(v_condition,' LIKE ');
      exit when v_loc = 0 ;
      v_beg := instr(substr(v_condition,1,v_loc-1),' ',-1) + 1;
      v_length := v_loc + 5 + instr(substr(v_condition,v_loc+6),' ')-v_beg;
      v_single_condition := substr(v_condition,v_beg,v_length);
      v_method := substr(v_single_condition,1,instr(v_single_condition,' ')-1);
      if v_method = 'NOT' then
        v_operand := 'NOT LIKE';
        v_loc := instr(v_condition,' NOT LIKE ');
        v_beg := instr(substr(v_condition,1,v_loc-1),' ',-1) + 1;
        v_length := v_loc + 9 + instr(substr(v_condition,v_loc+10),' ')-v_beg;
        v_single_condition := substr(v_condition,v_beg,v_length);
        v_method := substr(v_single_condition,1,instr(v_single_condition,' ')-1);
        v_value  := substr(v_single_condition,instr(v_single_condition,' NOT LIKE ')+10);
      else
        v_operand := 'LIKE';
        v_value  := substr(v_single_condition,instr(v_single_condition,' LIKE ')+6);
      end if;

      begin
        if v_method = 'ROLE' then
          if v_operand = 'LIKE' then
            select max(ans) into v_check from
	         (select 'TRUE' ans from dba_role_privs where grantee = v_username and granted_role like (v_value)
	          union
	          select 'FALSE' ans from dual);
	        else
	          select min(ans) into v_check from
	          (select 'FALSE' ans from dba_role_privs where grantee = v_username and granted_role like (v_value)
	           union
             select 'TRUE' ans from dual);
          end if;
        else  
        if v_method = 'PROGRAM' then
          if v_operand = 'LIKE' then
            select 'TRUE' into v_check from dual where upper(v_program) like (v_value);
          else
            select 'TRUE' into v_check from dual where upper(v_program) NOT like (v_value);
          end if;
        else
        if v_method = 'MODULE' then
          if v_operand = 'LIKE' then
            select 'TRUE' into v_check from dual where upper(v_module) like (v_value);
          else
            select 'TRUE' into v_check from dual where upper(v_module) NOT like (v_value);
          end if;
        else
        if v_method = 'USERNAME' then
          if v_operand = 'LIKE' then
            select 'TRUE' into v_check from dual where upper(v_username) like (v_value);
          else
            select 'TRUE' into v_check from dual where upper(v_username) NOT like (v_value);
          end if;
        else
        if v_method = 'OSUSER' then
          if v_operand = 'LIKE' then
            select 'TRUE' into v_check from dual where upper(v_osuser) like (v_value);
          else
            select 'TRUE' into v_check from dual where upper(v_osuser) NOT like (v_value);
          end if;
        else
        if v_method = 'LINK' then
          if v_operand = 'LIKE' then
            select 'TRUE' into v_check from dual where upper(v_program) like (v_value);
          else
            select 'TRUE' into v_check from dual where upper(v_program) NOT like (v_value);
          end if;
        else
        if v_method = 'MACHINE' then
          if v_operand = 'LIKE' then
            select 'TRUE' into v_check from dual where upper(v_machine) like (v_value);
          else
            select 'TRUE' into v_check from dual where upper(v_machine) NOT like (v_value);
          end if;
        else
          v_check := 'ERROR';
        end if;
        end if;
        end if;
        end if;
        end if;
        end if;
        end if;
      exception when no_data_found then
        v_check := 'FALSE';
      end;

      if v_check = 'ERROR' then
        v_error_condition := replace(v_error_condition,lower(v_single_condition),upper(v_single_condition));
      else
        v_condition := replace(v_condition,v_single_condition,v_check);
      end if;

      v_count := v_count + 1;
    end loop;  -- Step through each clause in the condition

    -- Check if there is Not an ERROR anywhere in the condition
    if instr(v_error_condition,'ERROR') = 0 then
      -- No-Error

      -- Check for mutiple boolean clauses in the condition
      if v_condition <> 'TRUE' and v_condition <> 'FALSE' then
        -- Evaluate the condition with a PL/SQL block to combine all of the clauses
        -- Sample 'TRUE and (TRUE or FALSE)' ==> TRUE
        v_sql := 'begin '||chr(10)||
                 'if '||v_condition||' then'||chr(10)||
                   'select '||CHR(39)||'TRUE'||CHR(39)||' into :xxx from dual;'||chr(10)||
                 'else'||chr(10)||
                   'select '||CHR(39)||'FALSE'||CHR(39)||' into :xxx from dual;'||chr(10)||
                 'end if;'||chr(10)||
                 'end;' ;
        begin
          execute immediate v_sql using out v_check;
        exception when others then
          v_check := 'FALSE';
          v_error_condition := 'ERROR: '||v_error_condition;
        end;
      end if;

      if v_check = 'TRUE' then
        if auth_control.permission = 'UNAUTHORIZED' then
          v_unauthorized_condition      := auth_control.condition;
          v_unauthorized_condition_eval := v_condition;
          v_unauthorized_level          := auth_control.permission_level;
          v_unauthorized_action         := auth_control.action;
          v_unauthorized_email_notice   := auth_control.email_notice;
          v_block_msg                   := auth_control.block_msg;
        else
          v_authorized_condition        := auth_control.condition;
          v_authorized_condition_eval   := v_condition;
          v_authorized_level            := auth_control.permission_level;
          v_authorized_action           := auth_control.action;
          v_authorized_email_notice     := auth_control.email_notice;
        end if;

        -- If any condition with action = 'LOG' or 'BLOCK' or 'TRACE' or 'RECORD' is met, then v_log = 'YES'
        if instr('LOG BLOCK TRACE RECORD',auth_control.action) > 0 then
          v_log             := 'YES';
        end if;

        -- If any condition with email_notice = 'S' is met, then v_sticky_email = 'YES'
        if auth_control.email_notice = 'S' then
          v_sticky_email           := 'YES';
        end if;

        -- If any condition with action = 'TRACE' or 'RECORD' is met then v_trace = 'TRACE' or 'RECORD'
        -- and applied provided the session is not Blocked.  'TRACE' is given precedince over 'RECORD'
        if auth_control.action = 'RECORD' and v_trace <> 'TRACE' then
          v_trace := 'RECORD';
        else
          if auth_control.action = 'TRACE' then
            v_trace := 'TRACE';
          end if;
        end if;

      end if; -- v_check = TRUE

    end if; -- No Error

  end loop;  -- Step throught each enabled condition ordered by permission_level asc

  -- Test for ERROR
  if instr(v_error_condition,'ERROR') > 0 then
    v_action := 'ERROR';
    v_log := 'YES';
    v_condition_eval := v_error_condition;
  else
    -- Determine if Authorized or Unauthorized.
    -- Authorized condition must be greater level than any Unauthorized condition level.
    if v_authorized_level > v_unauthorized_level then
      -- Authorized
      -- Possible v_action values: NONE, LOG, ERROR
      v_action         := v_authorized_action;
      v_condition      := v_authorized_condition;
      v_condition_eval := v_authorized_condition_eval;
      v_email          := v_authorized_email_notice;
    else
      -- Unauthorized
      -- Possible v_action values: NONE, LOG, BLOCK, ERROR
      v_action := v_unauthorized_action;
      v_condition      := v_unauthorized_condition;
      v_condition_eval := v_unauthorized_condition_eval;
      v_email          := v_unauthorized_email_notice;
    end if;
  end if;

  if v_log='YES' then
    -- If any condition with action = 'LOG' or 'BLOCK' or 'TRACE' or 'RECORD' or 'ERROR' was met, then v_log = 'YES'
    select auth.auth_log_id_seq.nextval into v_log_id from dual;
    insert into auth.auth_log
      (connection_date, log_id, username, dba, link, osuser, program, module, machine, action, condition, condition_eval)
    values (sysdate, v_log_id, v_username, v_dba, null, v_osuser, v_program, v_module, v_machine, v_action, v_condition, v_condition_eval );
    commit;
  end if;

  if v_sticky_email = 'YES' or v_email = 'Y' then
    -- Send email notification
    v_body := 'Authentication Notice'||chr(10)||
              'Date: '||to_char(sysdate,'YYYY-MM-DD HH:MI AM')||chr(10)||
              'Host Box: '||v_host_name||chr(10)||
              'Username: '||v_username||chr(10)||
              'DBA: '||v_dba||chr(10)||
              'OS User: '||v_osuser||chr(10)||
              'Program: '||v_program||chr(10)||
              'Module: '||v_module||chr(10)||
              'Machine: '||v_machine||chr(10)||
              'Action: '||v_action||chr(10)||
              'Condition: '||v_condition||chr(10)||
              'Condition Eval: '||v_condition_eval ;
    for email in (select email_address from auth.auth_email where upper(enabled) = 'Y') loop
      auth.auth_sendmail('Auth@'||v_host_name, email.email_address,'Auth Notice for '||v_username||' on '||v_host_name||':'||v_instance_name, v_body, null);
    end loop;
  end if;

  if v_action = 'BLOCK' then
    raise_application_error (-20001,v_block_msg,true);
  end if;

  if v_trace = 'RECORD' or v_trace = 'TRACE' then
    -- Get the System Process ID
    select p.spid, s.logon_time, s.serial# into v_spid, v_logon_time, v_serial
      from v$session s, v$process p where s.paddr = p.addr and s.sid = v_sid;
    -- Get trace file path and name
    select value||'/'||lower(v_instance_name)||'_ora_'||v_spid||'.trc' into v_trace_path_file
      from v$parameter where name = 'user_dump_dest';
    insert into auth.auth_trace (log_id, logon_time, sid, spid, trace_path_file, action, status)
      values (v_log_id, v_logon_time, v_sid, v_spid, v_trace_path_file,  v_trace ,'IN-PROGRESS');
    commit;
    -- Turn Tracing on for all sql processed by the session.
    -- execute immediate 'alter session set sql_trace = true';
    begin
      sys.dbms_system.set_sql_trace_in_session(v_sid,v_serial,true);
    exception when others then
      -- Try a different apporach
      begin
        execute immediate 'alter session set sql_trace = true';
      exception when others then
        raise_application_error (-20001,'Access Denied '||SQLCODE||' prevented trace',true);
      end;
    end;

    -- MAX_DUMP_FILE_SIZE specifies the maximum size of trace files (excluding the
    -- alert file). Change this limit if you are concerned that trace files may use too much
    -- space. Range of values 0 to unlimited, or UNLIMITED
    -- Requires "alter session" priv, but successful execution is Optional
    begin
      execute immediate 'alter session set MAX_DUMP_FILE_SIZE = UNLIMITED';
      exception when others then
        null;
    end;

  end if;

end;
/

grant execute on AUTH.AUTH_CONNECTION to SYS;
