#!/bin/ksh
#
# File: auth_trace.sh
# Works in concert with auth.auth_connnection procedure which may start a SQL trace session on a connection to an 
# instance bepending on the configuration in the auth.auth_control table.  The auth_trace.sh script checks if the
# traced session has completed, and if so it calls tkprof to covert the file to both output and a recording and then
# emails the results. The auth.auth_email table has a limit column and the value can be set to the username that was
# traced or to an control action keywork such as LOG, BLOCK, RECORD or TRACE. If a username is specified in the limit 
# column the only connections for the specified user will be emailed. If RECORD is specified only traces with the 
# RECORD action will be sent, TRACE will limit to TRACE actions and so on.
#
# Explain plans are generated if the auth_trace_sid.txt contains a user and password for the explain.
# The instances processed by auth_trace.sh must be in the auth_trace_sid.txt file or the file
# can contain a line with <ALL> to indicate that all instances should be processed. The script also deletes old 
# records from the auth_log and auth_trace tables and removes trace files, output files, and record files from the
# udummp directory when passed a parameter of -cleanup. The number of days of data retailed is set in the configuration
# section of the script. An email is sent to a list of recipients configured by the ERROR_EMAIL_LIST when an error is
# detected or when the verbose parameter (-v) is passed.
#
# Author: Doug Jackson  10/18/2004
#  Revised: 12/8/2005-Minor Changes
#  Revised: 03/31/2006-Added summary of long processes to the bottom of Trace results.
#  Revised: 04/18/2006-Added ability to limit emails based on username and ACTION.
#  Revised: 05/13/2009-Added RECORD traces are loaded to auth.auth_record and auth.auth_record_source in AUTH_RECORD_DB.
#  Revised: 01/11/2011-Added traps to igore errors caused by the database shutting down.

#--------------------------------------------------------------------
#- Send Error Mail
#--------------------------------------------------------------------
proc_send_error_mail()
{
  if [ "$STATUS" = "ERROR" ] || [ "$VERBOSE" = "YES" ]
  then
    # Send to error recipients 
    for EACH in `cat $ERROR_EMAIL_LIST | grep -v "^#" | sort | uniq`
    do
      mail -s "$STATUS: Running auth_trace.sh on $HOST" $EACH < $LOG
    done
  fi
}

#--------------------------------------------------------------------
#- Clean Up
#--------------------------------------------------------------------
proc_clean_up()
{
  if [ -f $TEMPFILE1 ] ; then rm -f $TEMPFILE1 ; fi
  if [ -f $TEMPFILE2 ] ; then rm -f $TEMPFILE2 ; fi
  if [ -f $TEMPFILE3 ] ; then rm -f $TEMPFILE3 ; fi
  if [ -f $TEMPFILE4 ] ; then rm -f $TEMPFILE4 ; fi
}

#################################################################################################################
### MAIN PROGRAM
#################################################################################################################

SCRIPT_DIR=`dirname $0`
if [ "$SCRIPT_DIR" = "." ]
then
  SCRIPT_DIR=`pwd`
fi

#Configuration
SID_LIST=$SCRIPT_DIR/auth_trace_sid.txt
ERROR_EMAIL_LIST=$SCRIPT_DIR/auth_trace_email.txt
ORATAB=/etc/oratab
AUTH_USERNAME="auth"
AUTH_PASSWORD="authpass"
AUTH_RECORD_DB=""
LOG_DAYS_RETAINED=35
TRACE_DAYS_RETAINED=7
# Default Cleanup Setting
CLEANUP=YES

LOG=$SCRIPT_DIR/auth_trace.log
TEMPFILE1=$SCRIPT_DIR/auth_trace1.tmp
TEMPFILE2=$SCRIPT_DIR/auth_trace2.tmp
TEMPFILE3=$SCRIPT_DIR/auth_trace3.tmp
TEMPFILE4=$SCRIPT_DIR/auth_trace4.tmp
STATUS="Normal"
HOST=`hostname`

# Verify that auth_trace.sh is not already running
if [ "`ps -ef | grep auth_trace | grep -v grep | grep -v vi | wc -l`" -gt "1" ]
then
  echo "auth_trace.sh is already running.. Exiting"
  exit
fi

# Initialize the Log file
echo Host: $HOST > $LOG
echo auth_trace.sh $* >> $LOG
date >> $LOG

# Make the dir with the SQL script the current directory
cd $SCRIPT_DIR

# Check for oratab file existence
if [[ ! -f $ORATAB ]]
then
  echo "ERROR: $ORATAB DOES NOT EXIST! auth_trace.sh failed."
  echo "ERROR: $ORATAB DOES NOT EXIST! auth_trace.sh failed." >> $LOG
  STATUS="ERROR"
  proc_send_error_mail
  proc_clean_up
  exit 1
fi

if [ ! -f $ERROR_EMAIL_LIST ]
then
  echo "ERROR: $ERROR_EMAIL_LIST DOES NOT EXIST! auth_trace.sh failed."
  echo "ERROR: $ERROR_EMAIL_LIST DOES NOT EXIST! auth_trace.sh failed." >> $LOG
  STATUS="ERROR"
  proc_send_error_mail
  proc_clean_up
  exit 1
fi

if [ ! -f $SID_LIST ]
then
  echo "ERROR: $SID_LIST DOES NOT EXIST! auth_trace.sh failed."
  echo "ERROR: $SID_LIST DOES NOT EXIST! auth_trace.sh failed." >> $LOG
  STATUS="ERROR"
  proc_send_error_mail
  proc_clean_up
  exit 1
fi

if [ "$CLEANUP" = "YES" ] || [ ! -z "`echo $* | grep -i \"-cleanup\"`" ]
then
  CLEANUP=YES
else
  CLEANUP=NO
fi

if [ ! -z "`echo $* | grep -i '\-v'`" ]
then
  VERBOSE=YES
else
  VERBOSE=NO
fi 

SID=`echo $* | grep -iv \"\-v\" | grep -iv \"\-cleanup\"`

if [ ! -z "$SID" ]
then
  # Test Passed SID
  if [ -z "`grep \"^${SID}:\" $ORATAB`" ]
  then
    echo "ERROR: A database named $SID was not found."
    echo "ERROR: A database named $SID was not found." >> $LOG
    STATUS="ERROR"
    proc_send_error_mail
    proc_clean_up
    exit 1
  else
    SID_LISTV=$SID
  fi
else
  # Step through each SID in SID_LIST or all SID in /etc/oratab if SID_LIST contains "<ALL>"
  if [ ! -z "`grep '^<ALL>$' $SID_LIST`" ]
  then
    SID_LISTV=`cat $ORATAB | cut -d: -f1 | grep -v ^# | grep -v ^* | grep -v ^$ | grep -v lsnr`
  else
    SID_LISTV=`cat $SID_LIST | grep -v ^# | cut -d: -f1`
  fi
fi

# Source the Oracle users's profile to set the path
. ~/.profile noup > /dev/null
. set_lib.sh

# Tempfile4 is used for insert statements that will load any RECORD traces into the auth.auth_record
# and auth.auth_record_source tables in the database specified by AUTH_RECORD_DB
>$TEMPFILE4

for SID in $SID_LISTV
do

  echo "=========================================================================================" >> $LOG
  echo "Checking Instance: $SID" >> $LOG

  STATUS="NORMAL"

  # Set the environment
  export ORACLE_SID=$SID
  export ORAENV_ASK=NO
  . oraenv
  . set_lib.sh

  # The database must be up
  echo "Checking Database Status" >> $LOG
  if [ -z "`ps -ef | grep -v grep | grep ora_smon_$ORACLE_SID`" ]
  then
    echo "Database $ORACLE_SID was not up. skipping..." >> $LOG
    continue
  fi

  echo "Verifing AUTH Existance Password..." >> $LOG
  STATUS=`sqlplus -s /nolog << ENDIT
set pause off head off pagesize 0 feedback off linesize 200
connect $AUTH_USERNAME/$AUTH_PASSWORD 
exit
ENDIT`
  if [ ! -z "`echo $STATUS | grep ORA- `" ]
  then
    echo "AUTH schema not available." >> $LOG
    echo $STATUS >> $LOG
    STATUS="WARNING"
    continue
  else
    echo "AUTH schema user is valid." >> $LOG
  fi

  if [ "$CLEANUP" = "YES" ]
  then
    echo "Cleanup..." >> $LOG
    echo "Deleting auth_log and auth_trace records older than $LOG_DAYS_RETAINED days." >> $LOG
sqlplus $AUTH_USERNAME/$AUTH_PASSWORD  << ENDIT >>/dev/null
prompt delete auth.auth_trace where logon_time < sysdate-$LOG_DAYS_RETAINED;    
delete auth.auth_trace where logon_time < sysdate-$LOG_DAYS_RETAINED;    
prompt delete auth.auth_log where connection_date < sysdate-$TRACE_DAYS_RETAINED;    
delete auth.auth_log where connection_date < sysdate-$TRACE_DAYS_RETAINED;    
commit;
exit
ENDIT

    # Get udump path
    echo "Getting user_dump_dest..." >> $LOG
UDUMP_DEST=`sqlplus -s $AUTH_USERNAME/$AUTH_PASSWORD  << ENDIT
set pause off head off pagesize 0 feedback off linesize 200
select value from v\\$parameter where name='user_dump_dest';
exit
ENDIT`
    echo UDUMP_DEST=$UDUMP_DEST >> $LOG
    if [ ! -z "`echo $UDUMP_DEST | egrep '(ORA-01092|ORA-03113)'`" ]
    then
      # Ignore the error: ORA-03113: end-of-file on communication channel
      # Ignore the error: ORA-03113: ORACLE instance terminated.  
      echo "ABORTING because the database is shutting down." >>$LOG
      break
    fi
    cd $UDUMP_DEST
    if [ ! "$UDUMP_DEST" = "`pwd`" ]
    then
      echo "ERROR: Failed to locate $UDUMP_DEST" >> $LOG
      STATUS="ERROR"
      break
    fi
    echo "Deleting trace files, output files, and record files older than $TRACE_DAYS_RETAINED days..." >>$LOG
    find ${UDUMP_DEST}/*.trc -mtime +${TRACE_DAYS_RETAINED} >> $LOG 2>/dev/null
    find ${UDUMP_DEST}/*.trc.gz -mtime +${TRACE_DAYS_RETAINED} >> $LOG 2>/dev/null
    find ${UDUMP_DEST}/*.out -mtime +${TRACE_DAYS_RETAINED} >> $LOG 2>/dev/null  
    find ${UDUMP_DEST}/*.txt -mtime +${TRACE_DAYS_RETAINED} >> $LOG 2>/dev/null 
    find ${UDUMP_DEST}/*.trc -mtime +${TRACE_DAYS_RETAINED} -type f -exec rm -f {} \; >> $LOG 2>/dev/null
    find ${UDUMP_DEST}/*.trc.gz -mtime +${TRACE_DAYS_RETAINED} -type f -exec rm -f {} \; >> $LOG 2>/dev/null
    find ${UDUMP_DEST}/*.out -mtime +${TRACE_DAYS_RETAINED} -type f -exec rm -f {} \; >> $LOG 2>/dev/null
    find ${UDUMP_DEST}/*.txt -mtime +${TRACE_DAYS_RETAINED} -type f -exec rm -f {} \; >> $LOG 2>/dev/null
  
    # Make the dir with the SQL script the current directory
    cd $SCRIPT_DIR
  fi

  # Get a list of trace records that are no longer in process. 
  echo "Checking for IN-PROGRESS trace records which no longer have a connected session or READY records." >> $LOG
PROCESS_LIST=`sqlplus -s $AUTH_USERNAME/$AUTH_PASSWORD  << ENDIT | tee $TEMPFILE1
set pause off head off pagesize 0 feedback off linesize 200
select to_char(t.logon_time,'YYYY/MM/DD-HH24:MI-')||t.spid||'='||t.trace_path_file 
  from auth.auth_trace t  
  where t.status = 'READY'
        or
        (t.status = 'IN-PROGRESS'
         and not exists (select 'x' from v\\$session s, v\\$process p where s.paddr = p.addr
                          and to_char(s.logon_time,'YYYY/MM/DD-HH24:MI-')||
                              p.spid = to_char(t.logon_time,'YYYY/MM/DD-HH24:MI-')||t.spid)
        );
exit
ENDIT`
  if [ ! -z "`grep 'ORA-' $TEMPFILE1`" ]
  then
    if [ ! -z "`egrep '(ORA-03113|ORA-01092)' $TEMPFILE1`" ]
    then
      # Ignore the error: ORA-03113: end-of-file on communication channel
      # Ignore the error: ORA-01092: ORACLE instance terminated.
      echo "ABORTING because the database is shutting down." >>$LOG
      break
    fi
    echo "ERROR: Getting PROCESS_LIST in $ORACLE_SID." >> $LOG
    grep "ORA-" $TEMPFILE1 >> $LOG
    STATUS="ERROR"
    break
  fi
  
  if [ -z "$PROCESS_LIST" ] 
  then
    echo "Nothing found in $SID" >> $LOG
    # Contine to next SID
    continue
  fi 

  # Step throught each in the list.
  for PROCESS_LINE in `echo $PROCESS_LIST`
  do 
   
    echo "Processing: $PROCESS_LINE..." >> $LOG
    PROCESS_ID=`echo $PROCESS_LINE | cut -d= -f1`
    TRACE_FILE=`echo $PROCESS_LINE | cut -d= -f2`

    # Write out the header.
    echo "Writing the header... " >> $LOG
sqlplus -s $AUTH_USERNAME/$AUTH_PASSWORD  << ENDIT > $TEMPFILE1
set pause off head off pagesize 0 feedback off linesize 200
select
'Username: '||l.username||chr(10)||
'Login: '||to_char(l.connection_date,'MON-DD HH:MI AM')||chr(10)||
'DBA: '||l.dba||chr(10)||
'Link: '||nvl(l.link,'NO')||chr(10)||
'OS User: '||l.osuser||chr(10)||
'Program: '||l.program||chr(10)||
'Module: '||l.module||chr(10)||
'Machine: '||l.machine||chr(10)||
'Action: '||t.action||chr(10)||
'Condition: '||l.condition||chr(10)||
'Condition Eval: '||l.condition_eval||chr(10)||
'Trace File: '||t.trace_path_file
from auth.auth_log l, auth.auth_trace t
where l.log_id = t.log_id
and to_char(t.logon_time,'YYYY/MM/DD-HH24:MI-')||t.spid = '$PROCESS_ID';
exit
ENDIT

    # Parse out the ACTION
    ACTION=`grep "^Action: " $TEMPFILE1 | cut -d: -f2 | sed 's/ //g'`
    # echo DEBUG ACTION=$ACTION

    # Get email list
    echo "Getting email list from auth_email..." >> $LOG
EMAIL_LIST=`sqlplus -s $AUTH_USERNAME/$AUTH_PASSWORD  << ENDIT | tee $TEMPFILE2
set pause off head off pagesize 0 feedback off linesize 200
select email_address from auth.auth_email where enabled='Y' 
and ( limit is null 
      or upper(limit) = ( select l.username from auth.auth_log l, auth.auth_trace t 
                          where l.log_id = t.log_id
                          and to_char(t.logon_time,'YYYY/MM/DD-HH24:MI-')||t.spid = '$PROCESS_ID' ) 
      or upper(limit) = ( select t.action from auth.auth_trace t 
                          where to_char(t.logon_time,'YYYY/MM/DD-HH24:MI-')||t.spid = '$PROCESS_ID' ) 
    ) ;
exit
ENDIT`
    if [ ! -z "`grep 'ORA-' $TEMPFILE2`" ]
    then
      if [ ! -z "`egrep '(ORA-03113|ORA-01092)' $TEMPFILE2`" ]
      then
        # Ignore the error: ORA-03113: end-of-file on communication channel
        # Ignore the error: ORA-01092: ORACLE instance terminated.
        echo "ABORTING because the database is shutting down." >>$LOG
        break
      fi
      echo "ERROR: Getting EMAIL_LIST in $ORACLE_SID." >> $LOG
      grep "ORA-" $TEMPFILE2 >> $LOG
      STATUS="ERROR"
      break
    fi

    # echo DEBUG TRACE_FILE=$TRACE_FILE
    if [ ! -f ${TRACE_FILE} ] && [ ! -f ${TRACE_FILE}.gz ]
    then
      # Try it with a lower case file name.
      TRACE_FILE=`dirname $TRACE_FILE`/`basename $TRACE_FILE | tr [A-Z] [a-z]`
    fi
    SEND_EMAIL=YES
    if [ ! -f ${TRACE_FILE} ] && [ ! -f ${TRACE_FILE}.gz ]
    then 
      # echo DEBUG Trace File is Missing
      echo "Trace File is Missing" >> $TEMPFILE1
      SEND_EMAIL=NO
      cat $TEMPFILE1 > $TEMPFILE2
      MAX_LAPSE=""
    else
      # Run tkprof against the trace file
      # if there is a user and password in the SID_LIST file for this instance then an explain is attempted.
      EXPLAIN_USER=`grep ^${ORACLE_SID}: $SID_LIST | cut -d: -f2`
      if [ ! -z "$EXPLAIN_USER" ]
      then
        EXPLAIN="explain=${EXPLAIN_USER}"
        USER=`echo ${EXPLAIN_USER} | cut -d/ -f1`
        echo "Generating Explain Plans with user: $USER" >> $LOG
      else
        EXPLAIN=""
        echo "No explain plan user specified in $SID_LIST" >> $LOG
      fi
      echo "Running tkprof against the trace file." >> $LOG
      OUTPUT_FILE=`echo $TRACE_FILE | sed 's/.trc/.out/g'`
      RECORD_FILE=`echo $TRACE_FILE | sed 's/.trc/.txt/g'`
      if [ -f $OUTPUT_FILE ] ; then rm -f $OUTPUT_FILE ; fi
      if [ -f $RECORD_FILE ] ; then rm -f $RECORD_FILE ; fi
      # Check if the trace file is zipped and unzip it if it is.
      if [ -f ${TRACE_FILE}.gz ]
      then
        gunzip ${TRACE_FILE}.gz
      fi
      tkprof $TRACE_FILE  $OUTPUT_FILE record=$RECORD_FILE $EXPLAIN sys=no >> $LOG
      # echo DEBUG RECORD_FILE=$RECORD_FILE
      # Zip the trace file 
      gzip ${TRACE_FILE}
      if [ ! -f $OUTPUT_FILE ] || [ ! -f $RECORD_FILE ] 
      then
        STATUS=ERROR
        break
      fi
  
      if [ "$ACTION" = "RECORD" ] && [ ! -z "$AUTH_RECORD_DB" ]
      then
        cat $RECORD_FILE >> $TEMPFILE1
        # SQL recording are mainly done for security purposes so we also record the SQL to a database.
        # Parse the Header info that was written to TEMPFILE1 
        USERNAME_V=`grep "^Username: " $TEMPFILE1 | cut -d: -f2 | sed 's/^ //g'` 
        DATE_V=`grep "^Login: " $TEMPFILE1 | cut -d: -f2 | sed 's/^ //g'`
        OSUSER_V=`grep "^OS User: " $TEMPFILE1 | cut -d: -f2 | sed 's/^ //g'`
        PROGRAM_V=`grep "^Program: " $TEMPFILE1 | cut -d: -f2 | sed 's/^ //g'`
        # Step through the RECORD_FILE and write insert statements to TEMPFILE4 
        COMMIT_COUNT=0
        STATEMENT_NUM_V=1
        LINE_NUM_V=0
        SESSION_TABLES=""

        # Following Used for Extracting Tables and View Names used in this Session.
        unset SESSION_TABLES
        SESSION_TABLES_CNT=0
        # Following Used for Extracting Tables and View Names used for each SQL statement.
        LEVEL=0
        unset LOOKING
        LOOKING[$LEVEL]=NO
        unset WAIT_FOR_COMMA
        WAIT_FOR_COMMA[$LEVEL]=NO
        
        grep -v "^/$" $RECORD_FILE | while read LINE
        do
          # echo DEBUG LINE=$LINE
          COMMIT_COUNT=`expr $COMMIT_COUNT + 1`
          LINE_NUM_V=`expr $LINE_NUM_V + 1`

          # Parse the LINE word by word looking for Tables.
          UPPER_LINE=`echo $LINE |tr [a-z] [A-Z]|sed 's/;/ ; /g'|sed 's/(/ ( /g'|sed 's/)/ ) /g'|sed 's/,/ , /g'|sed 's/  / /g'|sed 's/,/, /g'`
          for WORD in `echo $UPPER_LINE`
          do
            if [ -z "`echo $WORD | tr -d =`" ] || [ ! -z "`echo $WORD | grep \*`" ]
            then
              # Skip it
              continue
            fi

            if [ "${LOOKING[$LEVEL]}" = "YES" ]
            then
              # Things that end a list of tables, ie: "FOR SELECT","(join) ON", etc
              if [ ! -z "`echo \" $WORD \" | egrep '( ; | ON | WHERE | FOR | OUTER | INNER | LEFT | RIGHT | FULL | USING | NATURAL | CROSS  )'`" ]
              then
                LOOKING[$LEVEL]=NO
                # echo DEBUG: OFF WORD = $WORD
              fi
            fi

            if [ -z "`echo $WORD | tr -d \(`" ]
            then
              if [ $LEVEL -ge 20 ]
              then
                echo "Max Levels Exceeded"
              else  
                LEVEL=`expr $LEVEL + 1 `
              fi
              LOOKING[$LEVEL]=NO
            fi
            if [ -z "`echo $WORD | tr -d \)`" ]
            then
              LEVEL=`expr $LEVEL - 1 `
              WAIT_FOR_COMMA[$LEVEL]=YES
            fi  
            # echo DEBUG: LEVEL=$LEVEL
  
            if [ "${LOOKING[$LEVEL]}" = "YES" ] && [ "${WAIT_FOR_COMMA[$LEVEL]}" = "NO" ]
            then
              if [ ! "$WORD" = "DUAL" ]
              then
                # echo "DEBUG: TABLE = $WORD  ############################"
                # Search the array
                ROW=0
                FOUND=NO
                while [ "$ROW" -lt "$SESSION_TABLES_CNT" ]
                do
                  ROW=`expr $ROW + 1 `
                  # echo DEBUG $ROW ${SESSION_TABLES[$ROW]}
                  TABLE=`echo ${SESSION_TABLES[$ROW]} | cut -d: -f1`
                  if [ "$TABLE" = "$WORD" ]
                  then
                    FOUND=YES
                    COUNT=`echo ${SESSION_TABLES[$ROW]} | cut -d: -f2`
                    COUNT=`expr $COUNT + 1 `
                    SESSION_TABLES[$ROW]=$TABLE:$COUNT
                    break
                  fi
                done
                if [ "$FOUND" = "NO" ]
                then
                  SESSION_TABLES_CNT=`expr $SESSION_TABLES_CNT + 1 `
                  SESSION_TABLES[$SESSION_TABLES_CNT]=${WORD}:1
                fi
              fi  
              WAIT_FOR_COMMA[$LEVEL]=YES
            fi
 
            if [ -z "`echo $WORD | tr -d ,`" ] 
            then
              WAIT_FOR_COMMA[$LEVEL]=NO
            fi
            
            if [ "${LOOKING[$LEVEL]}" = "NO" ]
            then
              # Things that start a list of tables, ie: "FROM" and "JOIN (on)"
              if [ ! -z "`echo \" $WORD \" | egrep '( FROM | JOIN )'`" ]
              then
                LOOKING[$LEVEL]=YES
                WAIT_FOR_COMMA[$LEVEL]=NO 
                # echo DEBUG ON  WORD = $WORD
              fi
            fi
          done

          # Ready Line for insert 
          LINE=`echo "$LINE" | sed s/\'/\'\'/g` 
          echo "insert into auth.auth_record values ('$SID','$USERNAME_V',to_date('$DATE_V','MON-DD HH:MI AM')," >>$TEMPFILE4
          echo "'$OSUSER_V','$PROGRAM_V',$STATEMENT_NUM_V,$LINE_NUM_V,'$LINE');" >>$TEMPFILE4
          if [ $COMMIT_COUNT = 500 ]
          then
            echo "commit;" >>$TEMPFILE4
            COMMIT_COUNT=0
          fi

          if [ ! -z "`echo $LINE | grep \";$\"`" ]
          then
            # Next Statement
            STATEMENT_NUM_V=`expr $STATEMENT_NUM_V + 1`
            LINE_NUM_V=0
            
            # Take this opertunity to reset the variables used for parsing a statement in case we got out of wack.
            LEVEL=0
            unset LOOKING
            LOOKING[$LEVEL]=NO
            unset WAIT_FOR_COMMA
            WAIT_FOR_COMMA[$LEVEL]=NO
          fi

        done 
        
        # Write out insert lines to auth.auth_record_source
        # echo DEBUG SESSION_TABLES_CNT=$SESSION_TABLES_CNT
        ROW=0
        while [ "$ROW" -lt "$SESSION_TABLES_CNT" ]
        do
          ROW=`expr $ROW + 1 `
          TABLE=`echo ${SESSION_TABLES[$ROW]} | cut -d: -f1`
          COUNT=`echo ${SESSION_TABLES[$ROW]} | cut -d: -f2`
          echo "insert into auth.auth_record_source values ('$SID','$USERNAME_V'," >>$TEMPFILE4
          echo "to_date('$DATE_V','MON-DD HH:MI AM'),'$TABLE',$COUNT);" >>$TEMPFILE4
        done

        MAX_LAPSE=""

      else  # Not RECORD
 
        echo "Generating Explain Plans with user: $USER" >> $TEMPFILE1
        cat $OUTPUT_FILE | sed 's/^call/#call/g' | sed 's/^BEGIN/#BEGIN/g' | sed 's/^begin/#begin/g'  >> $TEMPFILE1
  
        # Find the top 10 long running SQL
        cat -n $TEMPFILE1 > $TEMPFILE2
        # OVERALL TOTALS FOR ALL NON-RECURSIVE STATEMENTS 
        PROCESS_LAPSE=`grep total $TEMPFILE2 | tail -2 | head -1 | awk -F" " '{print $5}'`
       
        # Count the total lines
        SQL_COUNT=`grep total $TEMPFILE2 | wc -l`
        # Less the summary totals
        SQL_COUNT=`expr $SQL_COUNT - 2 `
  
        ERR_COUNT=0
        > $TEMPFILE3
        grep total $TEMPFILE2 | head -$SQL_COUNT | while read LINE
        do
          if [ "`echo $LINE | awk -F\" \" '{print $2}'`" != "total" ]
          then
            ERR_COUNT=`expr $ERR_COUNT + 1 `
            continue
          fi
          echo $LINE | awk -F" " '{print $5, $1}' >> $TEMPFILE3
        done
        SQL_COUNT=`expr $SQL_COUNT - $ERR_COUNT `
  
        # Find Longest Lapse
        MAX_LAPSE=`cat $TEMPFILE3 | sort -n | tail -1 | cut -d' ' -f1`
        
        #Pull out the Top 10
        echo "***********************************************************************************************" >> $TEMPFILE1
        echo "***********************************************************************************************" >> $TEMPFILE1
        echo "**** TOP 10 LONGEST RUNNING SQL PROCESSES " >> $TEMPFILE1
        echo "**** Maximum Time: $MAX_LAPSE " >> $TEMPFILE1
        echo "***********************************************************************************************" >> $TEMPFILE1
        for LINE_NUM in `cat $TEMPFILE3 | sort -nr | head -10 | cut -d' ' -f2`
        do
          echo "==============================================================================================" >> $TEMPFILE1
          END_LINE=`tail +$LINE_NUM $TEMPFILE2 | grep "\*\*\*\*\*\*\*\*\*\*\*\*\*\*\*\*" | sort -r -n | tail -1 | awk -F" " '{print $1}'`
          BEG_LINE=`head -$LINE_NUM $TEMPFILE2 | sort -r | grep "\*\*\*\*\*\*\*\*\*\*\*\*\*\*\*\*" | sort -n | tail -1  | awk -F" " '{print $1}'`
          NUM_LINES=`expr $END_LINE - $BEG_LINE + 1 `
          head -n $END_LINE $TEMPFILE2 | tail -$NUM_LINES >> $TEMPFILE1
        done
      fi
    fi 
    if [ "$SEND_EMAIL" = "YES" ]
    then
      echo "Sending email..." >> $LOG
      for EACH in `echo $EMAIL_LIST`
      do
        mail -s "Auth Trace Results from $ORACLE_SID $MAX_LAPSE" $EACH < $TEMPFILE1 > $TEMPFILE2
        sleep 2
        if [ "`cat $TEMPFILE2 | wc -l`" -ne 0 ] 
        then
          echo "ERROR sending mail." >> $LOG
          cat $TEMPFILE2 >> $LOG
          STATUS="ERROR"
        fi
      done
    fi

    if [ ! "$STATUS" = "ERROR" ]
    then
      echo "Changed status from IN-PROGRESS to COMPLETED..." >> $LOG
      # Change the record from "IN-PROGRESS" to "COMPLETED'
sqlplus -s $AUTH_USERNAME/$AUTH_PASSWORD  << ENDIT > $TEMPFILE1
set pause off head off pagesize 0 feedback off linesize 200
update auth.auth_trace set status = 'COMPLETED' 
  where to_char(logon_time,'YYYY/MM/DD-HH24:MI-')||spid = '$PROCESS_ID';    
commit;
exit
ENDIT
    fi

  # End of trace file loop
  done

# End SID Loop
done

# If there were any RECORD action then the SQL has been written to TEMPFILE4 so it can be loaded to AUTH_RECORD_DB
if [ -s "$TEMPFILE4" ]
then
  echo "" >>$LOG
  echo "Loading RECORD action SQL to auth.auth_record in $AUTH_RECORD_DB." >> $LOG
  sqlplus -s $AUTH_USERNAME/$AUTH_PASSWORD@$AUTH_RECORD_DB  << ENDIT > $TEMPFILE1
set pause off head off pagesize 0 feedback off linesize 200
@$TEMPFILE4
commit;
exit
ENDIT
  cat $TEMPFILE1 > $SCRIPT_DIR/auth_trace_last_upload.sql
  if [ ! -z "`grep ORA- $TEMPFILE1`" ]
  then
    if [ ! -z "`egrep '(ORA-01092|ORA-03113)' $TEMPFILE1`" ]
    then
      # Ignore the error: ORA-03113: end-of-file on communication channel
      # Ignore the error: ORA-01092: ORACLE instance terminated. 
      echo "Failed to load RECORD data because the database is shutting down." >>$LOG
    else
      echo "ERROR: Failed to load RECORD data to auth.auth_record in $AUTH_RECORD_DB." >> $LOG
      grep ORA- $TEMPFILE1 >> $LOG
      STATUS="ERROR"
    fi 
  fi
fi

echo "Completing Cleanup..." >> $LOG
proc_send_error_mail
proc_clean_up
exit 0
# end of command file
